@@ -1,13 +1,13 @@
 """This module contains tools for describing objects being published to DLHub."""
 import importlib
-from itertools import zip_longest
-from datetime import datetime
-
+from typing import List, Sequence, Optional, Iterable, Union, Dict
 import pkg_resources
 import requests
+from enum import Enum
+
+from pydantic import BaseModel, Field
 from six import string_types
 from zipfile import ZipFile
-from copy import deepcopy
 from glob import glob
 import fnmatch
 import json
@@ -15,13 +15,41 @@ import sys
 import os
 import re
 
-from dlhub_sdk.version import __version__
-from dlhub_sdk.utils.schemas import codemeta_to_datacite
+from dlhub_sdk.models.datacite import Datacite, DataciteRelatedIdentifierType, DataciteRelationType
+from dlhub_sdk import __version__
 
 name_re = re.compile(r'^\S+$')
 
 
-class BaseMetadataModel:
+class DLHubType(Enum):
+    """Type supported by DLHub"""
+
+    servable = "servable"
+    dataset = "dataset"
+    pipeline = "piepline"
+
+
+class DLHubMetadata(BaseModel):
+    """Basic metadata for a DLHub artefact
+
+    Includes information used by the DLHub web service to recognize an object,
+    define how to build the computational environment,
+    control access to it, et cetera"""
+
+    version: str = Field(__version__, description="Version of the SDK used to create this object")
+    domains: List[str] = Field(default_factory=list, description="List of scientific domains that describe this object")
+    visible_to: List[str] = Field(['public'],
+                                  description="List of groups or identities for which this object is visible")
+    name: Optional[str] = Field(None, description="Name of the object", regex=r'^\S+$')
+    files: Dict[str, Union[str, List[str]]] = Field(default_factory=dict,
+                                                    description="List of files associated with this object")
+    dependencies: Dict[str, Dict[str, str]] = Field(default_factory=dict,
+                                                    description="Library dependencies for an object")
+    test: Optional[bool] = Field(None, description="Internal use only. Used for testing the DLHub service")
+    type: DLHubType = Field(DLHubType.servable, description="Type of this object")
+
+
+class BaseMetadataModel(BaseModel):
     """Base class for models describing objects published via DLHub
 
     Covers information that goes in the :code:`datacite` block of the metadata file and
@@ -53,46 +81,21 @@ class BaseMetadataModel:
     resubmitting a object description.
     """
 
-    def __init__(self):
-        """
-        Initialize a dataset record
-        """
-
-        # Populate initial fields
-        self._output = {"datacite": {
-            "creators": [],
-            "titles": [{'title': None}],
-            "publisher": "DLHub",
-            "publicationYear": str(datetime.now().year),
-            "identifier": {'identifier': '10.YET/UNASSIGNED', 'identifierType': 'DOI'},
-            "descriptions": [],
-            "fundingReferences": [],
-            "relatedIdentifiers": [],
-            "alternateIdentifiers": [],
-            "rightsList": []
-        }, "dlhub": {
-            "version": __version__,
-            "domains": [],
-            "visible_to": ["public"],
-            'name': None,
-            'files': {},
-        }}
-
-    def __getitem__(self, item):
-        return self._output[item]
+    dlhub: DLHubMetadata = Field(default_factory=DLHubMetadata, description="Metadata used by the DLHub web service")
+    datacite: Datacite = Field(default_factory=Datacite)
 
     @property
-    def name(self):
+    def name(self) -> str:
         """
         Get the name of the servable
 
         Returns:
             (string) Name of the servable
         """
-        return self['dlhub']['name']
+        return self.dlhub.name
 
     @classmethod
-    def create_model(cls, **kwargs):
+    def create_model(cls, **kwargs) -> 'BaseMetadataModel':
         """Instantiate the metadata model.
 
         Takes in arguments that allow metadata describing a dataset to be autogenerated. For
@@ -101,12 +104,11 @@ class BaseMetadataModel:
         """
         return cls()
 
-    def read_codemeta_file(self, directory=None):
+    def read_codemeta_file(self, directory: Optional[str] = None):
         """Read in metadata from a codemeta.json file
 
         Args:
-            directory (string): Path to directory contain the codemeta.json file
-                (default: current working directory)
+            directory (string): Path to directory contain the codemeta.json file (default: current working directory)
         """
 
         # If no directory, use the current
@@ -118,36 +120,27 @@ class BaseMetadataModel:
             codemeta = json.load(fp)
 
         # Convert it to datacite and store it
-        datacite = codemeta_to_datacite(codemeta)
-        self._output["datacite"].update(datacite)
+        self.datacite = Datacite.from_codemeta(codemeta)
 
         return self
 
-    def set_authors(self, authors, affiliations=list()):
-        """Add authors to a dataset
+    def set_creators(self, authors: List[str], affiliations: List[Sequence[str]] = ()) -> 'BaseMetadataModel':
+        """Add authors to this object
 
         Args:
-            authors ([string]): List of authors for the dataset.
-                In format: "<Family Name>, <Given Name>"
-            affiliations ([[string]]): List of affiliations for each author.
+            authors: List of authors for the dataset. In format: "<Family Name>, <Given Name>"
+            affiliations: List of affiliations for each author.
         """
-        for author, aff in zip_longest(authors, affiliations, fillvalue=[]):
-            # Get the authors
-            temp = author.split(",")
-            family = temp[0].strip()
-            given = temp[1].strip()
-
-            # Add them to the list
-            self._output["datacite"]["creators"].append({
-                "givenName": given,
-                "familyName": family,
-                "affiliations": aff,
-            })
+        self.datacite.set_creators(authors, affiliations)
         return self
 
-    def set_title(self, title):
-        """Add a title to the dataset"""
-        self._output["datacite"]["titles"][0]["title"] = title
+    def set_title(self, title: str) -> 'BaseMetadataModel':
+        """Set the title for this object
+
+        Args:
+            title: Desired title
+        """
+        self.datacite.set_title(title)
         return self
 
     def set_version(self, version):
@@ -156,62 +149,46 @@ class BaseMetadataModel:
         Args:
             version (string): Version number
         """
-        self._output["datacite"]["version"] = str(version)
+        self.datacite.version = str(version)
         return self
 
-    def set_abstract(self, abstract):
-        """Define an abstract for this artifact. Use for a high-level summary
+    def set_abstract(self, abstract: str) -> 'BaseMetadataModel':
+        """Define an abstract for this object. Use for a high-level summary
 
         Args:
-            abstract (string): Description of this artifact
+            abstract: Description of this artifact
         """
-        abs_block = {'description': abstract, 'descriptionType': 'Abstract'}
-
-        # Delete Abstract block if it already exists
-        self._output["datacite"]["descriptions"] = [
-            x for x in self._output["datacite"]["descriptions"]
-            if x["descriptionType"] != "Abstract"
-        ]
-
-        # Add the new block
-        self._output["datacite"]["descriptions"].append(abs_block)
+        self.datacite.set_abstract(abstract)
         return self
 
-    def set_methods(self, methods):
-        """Define a methods section for this artifact. Use to describe any specific details
+    def set_methods(self, methods: str) -> 'BaseMetadataModel':
+        """Define a methods section for this object. Use to describe any specific details
         about how the dataset, model, etc was generated.
 
         Args:
             methods (str): Detailed method descriptions
         """
-        method_blcok = {'description': methods, 'descriptionType': 'Methods'}
-
-        # Delete Methods block if it already exists
-        self._output["datacite"]["descriptions"] = [
-            x for x in self._output["datacite"]["descriptions"]
-            if x["descriptionType"] != "Methods"
-        ]
-
-        self._output["datacite"]["descriptions"].append(method_blcok)
+        self.datacite.set_methods(methods)
         return self
 
-    def set_domains(self, domains):
-        """Set the field of science that is associated with this artifcat
+    def set_domains(self, domains: List[str]):
+        """Set the field of science that is associated with this object
 
         Args:
-            domains ([string]): Name of a fields of science (e.g., "materials science")
+            domains: Names of fields of science (e.g., "materials science")
         """
-        self._output["dlhub"]["domains"] = domains
+        self.dlhub.domains = domains.copy()
         return self
 
-    def set_visibility(self, users=None, groups=None):
-        """Set the list of people this artifact should be visible to.
+    def set_visibility(self, users: Optional[Iterable[str]] = None,
+                       groups: Optional[Iterable[str]] = None):
+        """Set the list of people this object should be visible to.
 
         By default, it will be visible to anyone (["public"]).
 
         Args:
-            users ([string]): List of GlobusAuth UUIDs of allowed users
-            groups ([string]): List of GlobusAuth UUIDs of allowed Globus groups
+            users: GlobusAuth UUIDs of allowed users
+            groups: GlobusAuth UUIDs of allowed Globus groups
         """
         visibilities = []
         if not (users or groups):
@@ -223,23 +200,23 @@ class BaseMetadataModel:
             if groups:
                 for group in groups:
                     visibilities.append('urn:globus:groups:id:' + group)
-        self._output["dlhub"]["visible_to"] = visibilities
+        self.dlhub.visible_to = visibilities
         return self
 
     def set_doi(self, doi):
         """Set the DOI of this object, if available
 
         This function is only for advanced usage. Most users of the toolbox will not
-        know the DOI before sending the doi in to DLHub.
+        know the DOI before sending the object in to DLHub.
 
         Args:
             doi (string): DOI of the object
         """
-        self._output["datacite"]["identifier"]["identifier"] = doi
+        self.datacite.set_identifier(doi)
         return self
 
     def set_name(self, name):
-        """Set the name of artifact.
+        """Set the name of the object.
 
         Should be something short, descriptive, and memorable
 
@@ -248,7 +225,7 @@ class BaseMetadataModel:
         """
         if name_re.match(name) is None:
             raise ValueError('Name cannot contain any whitespace')
-        self._output["dlhub"]["name"] = name
+        self.dlhub.name = name
         return self
 
     def set_publication_year(self, year):
@@ -259,7 +236,7 @@ class BaseMetadataModel:
         Args:
             year (string): Publication year
         """
-        self._output["datacite"]["publicationYear"] = str(year)
+        self.datacite.publicationYear = str(year)
         return self
 
     def add_requirement(self, library, version=None):
@@ -280,22 +257,20 @@ class BaseMetadataModel:
             try:
                 module = importlib.import_module(library)
                 version = module.__version__
-            except Exception:
+            except (AttributeError, ModuleNotFoundError):
                 version = pkg_resources.get_distribution(library).version
         elif version == "latest":
             pypi_req = requests.get('https://pypi.org/pypi/{}/json'.format(library))
             version = pypi_req.json()['info']['version']
 
         # Set the requirements
-        if "dependencies" not in self["dlhub"]:
-            self["dlhub"]["dependencies"] = {"python": {}}
-        self._output["dlhub"]["dependencies"]["python"][library] = version
+        if "python" not in self.dlhub.dependencies:
+            self.dlhub.dependencies["python"] = {}
+        self.dlhub.dependencies["python"][library] = version
         return self
 
     def add_requirements(self, requirements):
-        """Add a dictionary of requirements
-
-        Utility wrapper for `add_requirement`
+        """Add several Python library requirements
 
         Args:
             requirements (dict): Keys are names of library (str), values are the version
@@ -331,141 +306,42 @@ class BaseMetadataModel:
 
         return self
 
-    def add_rights(self, uri=None, rights=None):
-        """Any rights information for this resource. Provide a rights management statement for the
-        resource or reference a service providing such information. Include embargo information if
-        applicable. Use the complete title of a license and include
-        version information if applicable.
-
-        Args:
-            uri (string): URI of the rights
-            rights (string): Description of the rights
-        """
-        if uri is None and rights is None:
-            raise ValueError('You must defined either a URI or the rights')
-
-        # Make the rights
-        new_rights = {}
-        if uri is not None:
-            new_rights['rightsURI'] = uri
-        if new_rights is not None:
-            new_rights['rights'] = rights
-
-        # Add it to the list
-        self._output["datacite"]["rightsList"].append(new_rights)
-        return self
-
-    def add_funding_reference(self, name, identifier=None, identifier_type=None,
-                              award_number=None, award_title=None, award_uri=None):
-        """Add a funding source to the list of resources
-
-        Args:
-            name (string): Name of funding provider
-            identifier (string): Identifier (e.g., ISNI) of the funder
-            identifier_type (string): Type of the identifier (ISNI, GRID, Crossref Funder ID, Other)
-            award_number (string): Code assigned by the funder
-            award_title (string): Title of the award
-            award_uri (string): URI of the award
-        """
-
-        # Error checking
-        if identifier_type is None and identifier_type is not None:
-            raise ValueError('Identifier type must be provided. "Other" is an option')
-        if identifier_type not in ['ISNI', 'GRID', 'Crossref Funder ID', 'Other']:
-            raise ValueError('Identifier type not recognized')
-
-        # Format in datacite format
-        funder = {'funderName': name}
-        if identifier is not None:
-            funder['funderIdentifier'] = {'funderIdentifier': identifier,
-                                          'funderIdentifierType': identifier_type}
-        if award_number is not None:
-            funder['awardNumber'] = {'awardNumber': award_number}
-            if award_uri is not None:
-                funder['awardNumber']['awardURI'] = award_uri
-        if award_title is not None:
-            funder['awardTitle'] = award_title
-
-        # Append to the total list
-        self._output["datacite"]["fundingReferences"].append(funder)
-
-        return self
-
-    def add_alternate_identifier(self, identifier, identifier_type):
-        """Add an identifier of this artifact in another service
-
-        Args:
-            identifier (string): Identifier
-            identifier_type (string): Identifier type
-        """
-
-        self._output["datacite"]["alternateIdentifiers"].append({
-            'alternateIdentifier': identifier,
-            'alternateIdentifierType': identifier_type
-        })
-        return self
-
-    def add_related_identifier(self, identifier, identifier_type, relation_type):
-        """Add an identifier of an artifact that is related to this resource (e.g., a paper
-        that describes a dataset).
-
-        You must define both the identifier and how it relates to this resource. The possible types
-        of relations are listed in the
-        `documentation for datacite
-        <https://schema.datacite.org/meta/kernel-4.1/doc/DataCite-MetadataKernel_v4.1.pdf>_`
-        on Page 25. The most common ones used in DLHub will likely be:
+    def add_related_resource(self, identifier: str, identifier_type: Union[str, DataciteRelatedIdentifierType],
+                             relation_type: Union[str, DataciteRelationType]) -> 'BaseMetadataModel':
+        """Add a resource that is related to this object.
 
+        We use the DataCite to describe the relations. Common relation types for DLHub objects are:
             - "IsDescribedBy": For a paper that describes a dataset or model
             - "IsDocumentedBy": For the software documentation for a model
             - "IsDerviedFrom": For the database a training set was pulled from
             - "Requires": For any software libraries that are required for this module
 
         Args:
-            identifier (string): Identifier
-            identifier_type (string): Identifier type
-            relation_type (string): Relation between
+            identifier: Identifier
+            identifier_type: Identifier type
+            relation_type: Relation between this identifier and the object you are describing
         """
-
-        if identifier_type not in ["ARK", "arXiv", "bibcode", "DOI", "EAN13", "EISSN", "Handle",
-                                   "IGSN", "ISBN", "ISSN", "ISTC", "LISSN", "LSID", "PMID",
-                                   "PURL", "UPC", "URL", "URN"]:
-            raise ValueError('Unknown identifier type ({}).'.format(identifier_type))
-        if relation_type not in ["IsCitedBy", "Cites", "IsSupplementTo", "IsSupplementedBy",
-                                 "IsContinuedBy", "Continues", "IsNewVersionOf",
-                                 "IsPreviousVersionOf", "IsPartOf", "HasPart", "IsReferencedBy",
-                                 "References", "IsDocumentedBy", "Documents", "IsCompiledBy",
-                                 "Compiles", "IsVariantFormOf", "IsOriginalFormOf",
-                                 "IsIdenticalTo", "HasMetadata", "IsMetadataFor", "Reviews",
-                                 "IsReviewedBy", "IsDerivedFrom", "IsSourceOf", "IsDescribedBy",
-                                 "Describes", "Requires", "IsRequiredBy"]:
-            raise ValueError('Unknown relation type: ({})'.format(relation_type))
-
-        self._output["datacite"]["relatedIdentifiers"].append({
-            'relatedIdentifier': identifier,
-            'relatedIdentifierType': identifier_type,
-            'relationType': relation_type
-        })
-        return self
+        self.datacite.add_related_identifier(identifier, identifier_type, relation_type)
 
     def add_file(self, file, name=None):
-        """Add a file to the list of files to be distributed with the artifact
+        """Add a file to the list of files to be distributed with this object
 
         Args:
             file (string): Path to the file
             name (string): Optional. Name of the file, if it is a file that serves a specific
-                purpose in software based on this artifact (e.g., if this is a pickle file
-                of a scikit-learn model)
+                purpose (e.g., "pickle" if this is a pickle file of a scikit-learn model)
         """
 
         if name is None or name == "other":
-            if "other" not in self._output["dlhub"]["files"]:
-                self._output["dlhub"]["files"]["other"] = []
-            self._output["dlhub"]["files"]['other'].append(file)
+            if "other" not in self.dlhub.files:
+                self.dlhub.files["other"] = []
+            self.dlhub.files["other"].append(file)
         else:
-            self._output["dlhub"]["files"][name] = file
+            self.dlhub.files[name] = file
         return self
 
-    def add_directory(self, directory, include=(), exclude=(), recursive=False):
+    def add_directory(self, directory: str, include: Union[str, Iterable[str]] = (),
+                      exclude: Union[str, Iterable[str]] = (), recursive: bool = False):
         """Add all the files in a directory
 
         Args:
@@ -476,9 +352,9 @@ class BaseMetadataModel:
         """
 
         # Make sure include and exclude are lists
-        if isinstance(include, str):
+        if isinstance(include, string_types):
             include = [include]
-        if isinstance(exclude, str):
+        if isinstance(exclude, string_types):
             exclude = [exclude]
 
         # Get potential files
@@ -500,45 +376,34 @@ class BaseMetadataModel:
 
         return self.add_files(files)
 
-    def add_files(self, files):
+    def add_files(self, files: Iterable[str]):
         """Add files that should be distributed with this artifact.
 
         Args:
-            files ([string]): Paths of files that should be published
+            files: Paths of files that should be published
         """
-
-        # Type checking
-        if isinstance(files, str):
-            files = [files]
-
         # Add the files
         for file in files:
             self.add_file(file)
         return self
 
-    def to_dict(self, simplify_paths=False, save_class_data=False):
-        """Render the dataset to a JSON description
+    def to_dict(self, simplify_paths: bool = False):
+        """Render the object to a JSON-ready dictionary
 
         Args:
             simplify_paths (bool): Whether to simplify the paths of each file
-            save_class_data (bool): Whether to save data about the class
         Returns:
             (dict) A description of the dataset in a form suitable for download
         """
 
         # Check for required fields
-        if self._output["datacite"]["titles"][0]["title"] is None:
+        if len(self.datacite.titles) == 0:
             raise ValueError('Title must be specified. Use `set_title`')
-        if self._output["dlhub"]["name"] is None:
+        if self.name is None:
             raise ValueError('Name must be specified. Use `set_name`')
 
         # Make a copy of the output
-        out = deepcopy(self._output)
-
-        # Add the name of the class to the output, if desired
-        if save_class_data:
-            out['@class'] = '{}.{}'.format(self.__class__.__module__,
-                                           self.__class__.__name__)
+        out = self.dict(exclude_none=True).copy()
 
         # Prepare the files
         if simplify_paths:
@@ -554,27 +419,8 @@ class BaseMetadataModel:
 
             # Copy over the current files list
             out["dlhub"]["files"] = files
-
         return out
 
-    @classmethod
-    def from_dict(cls, data):
-        """Reconstitute class from dictionary
-
-        Args:
-            data (dict): Metadata for this class
-        """
-
-        # Create the object, overwrite data
-        output = cls()
-        output._output = deepcopy(data)
-
-        # Make sure the class information is removed
-        if '@class' in output._output:
-            del output._output['@class']
-
-        return output
-
     def list_files(self):
         """Provide a list of files associated with this artifact.
 
@@ -582,7 +428,7 @@ class BaseMetadataModel:
             ([string]) list of file paths"""
         # Gather a list of all the files
         output = []
-        for k, v in self._output["dlhub"]["files"].items():
+        for k, v in self.dlhub.files.items():
             if isinstance(v, string_types):
                 output.append(v)
             else:  # It is a list
